/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include "glView.h"

bool hlp_draw_grid_cross_hatch( rgVector *v, float res, GLView *g,
	const float min_scale, const float alpha, rgVector col,
	rgVector fcol, rgVector ocol);

void hlp_draw_grid_lines(
	rgVector a,
	rgVector b,
	const rgVector start,
	const rgVector stop,
	const rgVector nstep,	// Normalized step vector
	const rgVector step,
	const float res,
	const float fade,
	const float alpha,
	const rgVector col,
	const rgVector fcol,
	const rgVector ocol);

bool draw_axis_aligned_grid( GLView *g, float res, rgMatrix &itm,
	const float min_scale, const float alpha,
	rgVector col, rgVector fcol, rgVector ocol)
{
	if( res <= 0.0f) return false;

	// The grid bounding box is a clock-wise defined rectangle/square.
	// The first vertex is the top left corner (in window space).
	rgVector v[4] = {
		rgVector( 0, 0, 0),
		rgVector( g->width(), 0, 0),
		rgVector( g->width(), g->height(), 0),
		rgVector( 0, g->height(), 0)
	};

	// Convert from window space to world space.
	rggl_ortho_window_to_world( float( g->width()), float( g->height()),
		v, 4, &itm);

	// Draw the grid lines now.
	glBegin( GL_LINES);
	bool vis = hlp_draw_grid_cross_hatch( v, res, g, min_scale,
		alpha, col, fcol, ocol);
	glEnd();

	return vis;
}

bool hlp_draw_grid_cross_hatch( rgVector *v, float res, GLView *g,
	const float min_scale, const float alpha, rgVector col,
	rgVector fcol, rgVector ocol)
{
	// Increase the coarseness of the grid as it is scaled down. This
	// helps prevent drawing too many grid lines.
	//
	// NOTE: Only increase the coarseness on powers of two sizes
	//	(this prevents the grid from "jumping" around as it is
	//	resized).
	float fade;
	{
		fade = (v[1] - v[0]).length() / float( g->width());
		int k = int( fade);
		if( min_scale > (1.0f / fade)) return false;
		if( k >= 1) {
			int d = k;
			int mask = 1;
			while( d & (~0x1)) { d >>= 1; mask <<= 1;}
			k = k & mask;
			res *= float( k);
			

			fade = 1.0f - (fabs( fade - float( k)) / float( k));
		} else fade = 1.0f;
	}

	// Expand the grid bounding box so that it covers the entire window.
	const rgVector uvec = (v[1] - v[0]).normalize();
	const rgVector vvec = (v[3] - v[0]).normalize();
	const rgVector ustep( uvec * res);
	const rgVector vstep( vvec * res);
	v[0] = (v[0] - ustep) - vstep;
	v[1] = (v[1] + ustep) - vstep;
	v[2] = (v[2] + ustep) + vstep;
	v[3] = (v[3] - ustep) + vstep;

	// Align the world space grid bounding box to the given resolution.
	// NOTE: The grid is always fixed at (0, 0, 0). Also, round away from
	//	zero to avoid singularity (when bounding box rounds down to a
	//	point at the origin).
	{
		const float ires = 1.0f / res;
		for( int i = 0; i < 4; i++) {
			v[i].x = float( int( v[i].x * ires)) * res;
			v[i].y = float( int( v[i].y * ires)) * res;
			v[i].z = float( int( v[i].z * ires)) * res;
		}
	}

	// Draw horizontal grid lines.
	hlp_draw_grid_lines( v[0], v[1], v[0], v[3], vvec, vstep, res, fade,
		alpha, col, fcol, ocol);

	// Draw vertical grid lines.
	hlp_draw_grid_lines( v[3], v[0], v[0], v[1], uvec, ustep, res, fade,
		alpha, col, fcol, ocol);

	return true;
}

void hlp_draw_grid_lines(
	rgVector a,
	rgVector b,
	const rgVector start,
	const rgVector stop,
	const rgVector nstep,	// Normalized step vector
	const rgVector step,
	const float res,
	const float p_fade,
	const float alpha,
	const rgVector col,
	const rgVector fcol,
	const rgVector ocol)
{
	// Draw horizontal grid lines.
	const float fade = p_fade * alpha;
	int count = int( (stop - start).length() / res);
	bool do_fade = false;
	if( int( round( nstep.dot( a) / res)) & 0x1) do_fade = true;
	while( count--) {
		if( fabs( nstep.dot( a)) < (res * 0.5f))
			glColor4f( ocol.x, ocol.y, ocol.z, 1.0f);
		else {
			if( do_fade) {
				rgVector c = (col * fade) + fcol * (1.0f - fade);
				glColor4f( c.x, c.y, c.z, fade);
			}
			else glColor4f( col.x, col.y, col.z, alpha);
		}
		glVertex( a);
		glVertex( b);
		a = a + step;
		b = b + step;

		do_fade = !do_fade;
	}
}
