/*
Copyright (C) 2004-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#include "rg3DRom.h"
#include "rg3DRom_support.h"
#include "rg3DRom_csg.h"
#include "rgPort.h"
#include <iostream>
#include <fstream>
#include <string.h>
#include <rgError.h>
#include "rg3DRom_file.h"

int sk3DRomModel::load( const char *fname)
{
	const int err = load_3DRom( fname, *this);
	if( err) return err;

	check_faces();
	// Initialize the output skeleton (since it is undefined by default).
	skel.init_skel_frame( skel.skout);

	return 0;
}

int load_3DRom( const char *fname, sk3DRomModel &mdl)
{
	mdl.clear();

	if( !fname) return -1;
	std::ifstream inf( fname, std::ios::binary);

	// Read file identifier.
	sk3DRomHeader head;
	inf.read( head.header, 8);

	// Check the file identifier
	if( strncmp( head.header, sk3DROM_HEADER, 8) != 0) return -2;

	sk3DRomTag t;
	if( sk3DRom_read_tag( inf, t)) return -3;

	int seek_stop;
	switch( t.mode) {
	case sk3DROM_MODEB:
	case sk3DROM_MODEC:
		seek_stop = int( inf.tellg()) + t.size;
		while( inf.tellg() < seek_stop) {

			if( sk3DRom_read_tag( inf, t)) return -5;
			switch( t.mode) {
			case sk3DROM_TEXTURE:
				if( mdl.read_texture_data( inf)) return -6;
				break;

			case sk3DROM_FACE:
				if( mdl.read_face_data( inf)) return -7;
				break;

			case sk3DROM_VERTEX:
				if( mdl.read_vertex_data( inf)) return -9;
				break;

			// XXX: No longer supported.
			//case sk3DROM_VERTMODNAMES:
				//if( mdl.read_skeleton_names( inf, mdl.sn))
				//	return -10;
			//	break;

			case sk3DROM_SKELETON:
				if( mdl.read_skeleton( inf, mdl.skel))
					return -11;
				break;

			default:
				#ifdef LOAD_3DROM_DEBUG
				std::cerr << "*** seek ahead " << t.size <<
					" bytes\n";
				#endif

				inf.seekg( t.size, std::ios::cur);
				if( inf.fail()) return -1000;
				break;
			}
		}

		mdl.skel.build_master_list();

		// MODEB and MODEC 3DRoms do not have angle-based joints.
		// These must be generated automatically.
		mdl.auto_gen_joints();
		mdl.skel.bind_skeleton();

		// NOTE: Vertices must be mapped AFTER the skeleton is bound.
		mdl.map_verticies_to_skeleton();

		// MODEB and MODEC 3DRoms do not store master joint links.
		// These must be generated automatically.
		//mdl.gen_master_joint_links();
		mdl.build_bone_less_list();
		mdl.set_bid_for_all_verts();

		#ifdef LOAD_3DROM_DEBUG
		if( int( inf.tellg()) == seek_stop)
			std::cerr << "sk3DROM_MODEB/MODEC ok\n";
		#endif
		break;

	case sk3DROM_MODED:
		seek_stop = int( inf.tellg()) + t.size;
		while( inf.tellg() < seek_stop) {
			if( sk3DRom_read_tag( inf, t)) return -5;
			switch( t.mode) {
			case sk3DROM_MODEL_TEXTURES:
				if( mdl.read_model_texture_data( inf))
					return -6;
				break;
			
			case sk3DROM_CSG_DATA:
				if( mdl.read_csg_data( inf)) return -7;
				break;

			case sk3DROM_PAGES_DATA:
				if( mdl.read_pages_data( inf)) return -8;
				break;

			case sk3DROM_GROUPS_DATA:
				if( mdl.read_group_data( inf)) return -8;
				break;

			case sk3DROM_FACE:
				if( mdl.read_face_data( inf)) return -9;
				break;

			case sk3DROM_FACE_2:
				if( mdl.read_face_2_data( inf)) return -10;
				break;

			case sk3DROM_VERTEX:
				if( mdl.read_vertex_data( inf)) return -11;
				break;

			case sk3DROM_VERTEX_2:
				if( mdl.read_vertex_2_data( inf)) return -12;
				break;

			case sk3DROM_VERTEX_3:
				if( mdl.read_vertex_3_data( inf)) return -13;
				break;

			case sk3DROM_MAIN_SKELETON:
				if( mdl.read_main_skeleton_data( inf))
					return -20;
				break;

			case sk3DROM_SKELFRAMES_2:
				if( mdl.read_skel_frames_2_data( inf))
					return -30;
				break;

			case sk3DROM_SKEL_ANIMATION:
				if( mdl.read_skel_animation_data( inf))
					return -40;
				break;

			case sk3DROM_CUSTOM:
			// If there is a custom tag handler then read
			// the custom tag and allow the custom handler to
			// process it. Otherwise, save the custom tag into
			// memory so that it can be saved again later
			// (transparently).
			//
			// NOTE: It is the responsiblity of the custom tag
			//	custom handler to make sure that its custom
			//	data is consistant with the core data stored
			//	in official 3DRom tags.
			//
			// FIXME: A secure and efficient/simple way of letting
			//	custom tag handlers know when the core data
			//	has changed is required. Perhaps setting up
			//	flags to notify custom tag handlers when
			//	core data has changed.
				if( mdl.read_custom_data) {
					// Determine where the custom tag ends.
					const int final_seek =
						int( inf.tellg()) +
						t.size;

					// Skip the 64 byte identifier.
					inf.seekg( 64, std::ios::cur);
					if( inf.fail()) return -125;

					// Process the custom tag.
					const int err =
						mdl.read_custom_data(
						inf, &mdl, final_seek);
					if( err) return err;

					// If custom tag processing function
					// did not read all custom data, then
					// skip to the end of the tag. If it
					// read more data than was contained
					// in the tag, signal an error.
					if( inf.tellg() < final_seek) {
						inf.seekg( final_seek,
						std::ios::beg);
						if( inf.fail()) return -130;
					} else if( inf.tellg() > final_seek) {
						return -135;
					}

				// Save the custom tag into memory so that it
				// can be written out later when a save is
				// requested.
				} else {
					if( mdl.store_custom_tag_in_memory(
						inf, t.size)) return -140;
				}
				break;

			default:
				inf.seekg( t.size, std::ios::cur);
				if( inf.fail()) return -1000;
				break;
			}
		}

		// Some required data is not stored in a "MODED" file,
		// however it can be generated. Therefore generate all
		// required remaining data.
		mdl.skel.build_master_list();
		mdl.skel.gen_absolute_joint_verts();

		// FIXME: This should be removed as bone-less
		// information should be stored in the file.
		mdl.build_bone_less_list();
		mdl.set_bid_for_all_verts();
		mdl.build_group_vert_lists();	
		mdl.skel.build_root_bone_list();
		break;

	default:
		return -4;
		break;
	}

	#ifdef LOAD_3DROM_DEBUG
	std::cerr << "Final seek = " << inf.tellg() << "\n";
	#endif
	return 0;
}

int sk3DRomModel::store_custom_tag_in_memory( std::ifstream &inf,
	const int len)
{
	if( len <= 0) return -1;
	if( len > SK3DROM_MAXIMUM_CUSTOM_TAG_MEMORY_STORAGE) return -2;

	clear_custom_tag_memory();
	custom_tag_memory_length = len;
	custom_tag_memory_storage = new char[ len];
	if( !custom_tag_memory_storage) return -3;
	inf.read( custom_tag_memory_storage, len);
	if( inf.fail()) return -4;
	return 0;
}

void sk3DRomModel::clear_custom_tag_memory()
{
	custom_tag_memory_length = 0;
	if( custom_tag_memory_storage) {
		delete [] custom_tag_memory_storage;
		custom_tag_memory_storage = 0;
	}
}

int sk3DRomModel::get_custom_tag_memory_size()
{
	return custom_tag_memory_length;
}

int sk3DRomModel::read_texture_data( std::istream &inf)
{
	short count;
	inf.read( (char *)&count, 2);
	if( inf.fail()) return -1;

	#ifdef TEXTURE_DATA_DEBUG
	std::cerr << "TEXTURE\n";
	std::cerr << "\tcount = " << count << "\n";
	#endif

	while( count--) {

		// Read texture file name
		{
			short len;
			inf.read( (char *)&len, 2);

			#ifdef TEXTURE_DATA_DEBUG
			char *fname = new char[ len + 1];
			inf.read( fname, len);
			if( inf.fail()) return -3;
			fname[ len] = '\0';     // Null terminate string
			std::cerr << "\ttexture = \"" << fname << "\"\n";
			delete [] fname;

			#else
			inf.seekg( len, std::ios::cur);
			if( inf.fail()) return -3;
			#endif
		}
	}

	return 0;
}

int sk3DRomModel::read_skeleton_names( std::ifstream &inf,
	std::vector <char *> &sn)
{
        short count;
        inf.read( (char *)&count, 2);
        if( inf.fail()) return -1;
                                                                                
        #ifdef SKELETON_NAME_DEBUG
        std::cerr << "SKELETON MODEL NAMES (VERTEX MODEL NAMES)\n";
        std::cerr << "\tcount = " << count << "\n";
        #endif
                                                                                
                                                                                
        char *temp;
        short len;
        while( count--) {
                inf.read( (char *)&len, 2);
                if( inf.fail()) return -2;
                if( len < 0) return -3;
                                                                                
                temp = new char[ len + 1];
                if( !temp) return -4;
                if( len) {
                        inf.read( temp, len);
                        if( inf.fail()) return -5;
                }
                temp[ len] = '\0';      // Null terminate the string
                sn.push_back( temp);
                                                                                
                #ifdef SKELETON_NAME_DEBUG
                std::cerr << "\tname = \"" << temp << "\"\n";
                #endif
        }
                                                                                
        return 0;
}

int sk3DRomModel::read_skeleton( std::ifstream &inf, sk3DRomSkel &s)
{
        short count;
        inf.read( (char *)&count, 2);
        if( inf.fail()) return -1;
                                                                                
        #ifdef SKELETON_DEBUG
        std::cerr << "SKELETON MODEL\n";
        std::cerr << "\tcount = " << count << "\n";
        int bid = 0;
        #endif
                                                                                
        if( count < 0) return -2;
                                                                                
        // Read in the skeleton
        sk3DRomBone *b;
        while( count--) {
                s.b.push_back( sk3DRomBone());
                b = &(s.b[ s.b.size() - 1]);
                                                                                
                // Read bone joints
		b->a.read_binary( inf);
		b->b.read_binary( inf);
                b->a.y *= -1.0f;
                b->b.y *= -1.0f;
                                                                                
                // Initialize extra bone data
                b->pj = 0;
                                                                                
                // Read master bone index
		//inf.read( (char *)&(b->mid), 2);
		b->mid = read_word( inf);
                #ifdef SKELETON_DEBUG
                std::cerr << "(" << (bid++) << ")\tmaster = " << b->mid;
                #endif
                                                                                
                // Read all slave bones
                {
                        short scount;
                        inf.read( (char *)&scount, 2);
#ifdef SKELETON_DEBUG
                        std::cerr << "\t| #of slaves = " << scount;
                        #endif
                                                                                
                        if( scount < 0) return -7;
                                                                                
                        // Read all bone id's
                        sk3DRomBoneLink *link;
                        short jcount = scount;
                        while( scount--) {
                                b->link.push_back( sk3DRomBoneLink());
                                link = &(b->link[ b->link.size() - 1]);
                                inf.read( (char *)&(link->bid), 2);
                                if( inf.fail()) return -10;
                        }
                                                                                
                        // Read all joint id's
                        scount = 0;
                        while( jcount--) {
                                link = &(b->link[ scount++]);
                                inf.read( (char *)&(link->jid), 1);
                                if( inf.fail()) return -11;
                        }
                }
                                                                                
                // Read all controlled verticies
                {
                        short vcount;
                        inf.read( (char *)&vcount, 2);
                        #ifdef SKELETON_DEBUG
                        std::cerr << "\t| #of verts = " << vcount;
                        #endif
                        if( vcount < 0) return -12;
                                                                                
                        short vid;
                        while( vcount--) {
                                inf.read( (char *)&vid, 2);
                                if( inf.fail()) return -15;
                                b->vid.push_back( vid);
                        }
		}
                #ifdef SKELETON_DEBUG
                std::cerr << "\n";
                #endif
        }
                                                                                
        // Read in the skeleton key frames
        inf.read( (char *)&count, 2);
                                                                                
        #ifdef SKELETON_DEBUG
        std::cerr << "\n\ttotal skel key frames = " << count << "\n";
        #endif
        if( count < 0) return - 30;
                                                                                
        sk3DRomSkelFrame *sf;
        const short bcount = short( s.b.size());
        short i;
        while( count--) {
                s.sf.push_back( sk3DRomSkelFrame());
                sf = &(s.sf[ s.sf.size() - 1]);
                // FIXME: "reserve" does nothing!!!
                //sf->b.reserve( bcount);
                //std::cerr << "bcount = " << sf->b.size() << "\n";
                for( i = 0; i < bcount; i++) {
                        // Read in the 3 matrics (transformation, translation
                        // and rotation). Each matrix is 64 bytes (16 values of
                        // four bytes each).
                        sf->b.push_back( sk3DRomBoneFrame());
                        // Skip the 3 matricies since they are no longer
                        // being used (obsolete).
                        inf.seekg( 64 * 3, std::ios::cur);
                        //inf.read( (char *)((sf->b[i]).m.d), 64);
                        //inf.read( (char *)((sf->b[i]).t.d), 64);
//inf.read( (char *)((sf->b[i]).r.d), 64);
                        if( inf.fail()) return -40;
                }
        }
                                                                                
        return 0;
}

int sk3DRomModel::read_vertex_data( std::istream &inf)
{
        short scount;
        // Read in the number of pre-calculated vertices for each skeleton
        // frame.
        //
        // NOTE: This is an obsolete style which was intended to save loaders
        //      some time by not forcing them to calculate the vertex
        //      positions on their own at load time. However, from here on in,
        //      all loaders are expected to be fast enough to calculate their
        //      own vertices, so it is possible for a model file to have
        //      thousands of skeleton frames but only one pre-calculated
        //      vertex frame (there MUST be at least one frame, this is
        //      usually the frame were all bones are at neutral/default
        //      positions). Storing precalculated vertices for all skeleton
        //      frames is a waste of memory.
        inf.read( (char *)&scount, 2);
        if( inf.fail()) return -1;
	scount = rgPortLittleToNativeWORD( scount);
                                                                                
        #ifdef VERTEX_DEBUG
        std::cerr << "VERTEX\n";
        std::cerr << "\tskeleton count = " << scount << "\n";
        #endif

	if( scount < 0) return -2;
                                                                                
        short count;
        rgVector u;
        bool verts_read = false;
        while( scount--) {
                inf.read( (char *)&count, 2);
		count = rgPortLittleToNativeWORD( count);
                #ifdef VERTEX_DEBUG
                std::cerr << "\tvertex count = " << count << "\n";
                #endif
                if( count < 0) return -3;
                if( !verts_read) {
                        while( count--) {
                                v.push_back( u.read_binary( inf));
				{
					sk3DRomVertex &tmp = v[ v.size() - 1];
					tmp.m_gid = -1;
					tmp.flags = 0;
				}
                        }
                        verts_read = true;
		} else while( count--) u.read_binary( inf);
        }
                                                                                
        return 0;
}

int sk3DRomModel::read_vertex_2_data( std::istream &inf)
{
	short count = read_word( inf);	// Get the number of vertices
	#ifdef VERTEX_DEBUG
	std::cerr << "\tvertex count = " << count << "\n";
	#endif
	if( count < 0) return -3;

	// Read in all vertices.
	while( count--) {
		rgVector u;
		v.push_back( u.read_binary( inf));

		sk3DRomVertex &tmp = v[ v.size() - 1];
		// FIXME: Vertices should use a 32-bit gid in next version.
		tmp.m_gid = short( read_word( inf));
		tmp.flags = short( read_word( inf));
	}
	if( inf.fail()) return -4;
        return 0;
}

int sk3DRomModel::read_vertex_3_data( std::istream &inf)
{
	switch( read_dword( inf)) {
	case SK3DROM_VERTEX_VERSION_A:
		return read_version_A_vertex_data( inf); break;

	default: return -2;	// Currently not supporting unknown versions.
	}
}

int sk3DRomModel::read_version_A_vertex_data( std::istream &inf)
{
	int count = read_dword( inf);	// Get the number of vertices
	if( count < 0) return -3;

	// Read in all vertices.
	while( count--) {
		rgVector u;
		v.push_back( u.read_binary( inf));

		sk3DRomVertex &tmp = v[ v.size() - 1];
		tmp.weight = read_float( inf);
		tmp.m_gid = read_dword( inf);
		tmp.flags = read_dword( inf);
	}
	if( inf.fail()) return -4;
        return 0;
}

inline int hlp_custom_identifier_compare( const char *id, const char *buf)
{
	if( strncmp( id, buf, 64)) return 0;
	return 1;
}

int sk3DRom_read_and_compare_custom_identifier( std::ifstream &inf,
	const char *id)
{
	if( !id) return -1;

	char buf[64];
	inf.read( buf, 64);
	if( inf.fail()) return -2;

	return hlp_custom_identifier_compare( id, buf);
}

int sk3DRomModel::read_main_skeleton_data( std::ifstream &inf)
{
	// Read in the total number of bones.
	rgWORD n;
	inf.read( (char *)&n, 2);
	n = rgPortLittleToNativeWORD( n);
	if( inf.fail()) return -10;
	if( short( n) < 0) return -11;

	// Remove all current bone data (if any).
	skel.b.clear();

	// Read in all of the main skeleton's bones.
	sk3DRomTag t;
	int final_seek;
	const int istop = int( n);
	for( int i = 0; i < istop; i++) {
		
		// Read in the bone tag (which MUST be sk3DROM_BONE or
		// a later supported tag version).
		if( sk3DRom_read_tag( inf, t)) return -20;
		if( t.size < 0) return -22;
		final_seek = t.size + inf.tellg();	// End of bone data

		switch( t.mode) {
		case sk3DROM_BONE:
			read_bone_data( inf);
			break;

		case sk3DROM_BONE_2:
			read_bone_2_data( inf);
			break;

		default: return -21;
		}

		// If this bone tag contains more data then this function is
		// currently interested in, then skip the rest of the data.
		if( inf.tellg() < final_seek) {
			inf.seekg( final_seek, std::ios::beg);
			if( inf.fail()) return -500;
		}
	}


	return 0;
}

int sk3DRomModel::read_bone_2_data( std::ifstream &inf)
{
	// Create a new bone
	skel.b.push_back( sk3DRomBone());
	sk3DRomBone &b = skel.b[ skel.b.size() - 1];

	// Read bone version number
	rgWORD n;
	inf.read( (char *)&n, 2);
	n = rgPortLittleToNativeWORD( n);

	if( read_basic_bone_data( inf, b)) return -10;

	// Read in the bone data based on the version number.
	switch( n) {
	case SK3DROM_BONE_VERSION_A:
		inf.read( b.bname, SK3DROM_BONE_NAME_LEN);
		b.bname[ SK3DROM_BONE_NAME_LEN - 1] = '\0';
		if( inf.fail()) return -20;
		break;

	}

	return 0;
}

int sk3DRomModel::read_bone_data( std::ifstream &inf)
{
	// Create a new bone
	skel.b.push_back( sk3DRomBone());
	return read_basic_bone_data( inf, skel.b[ skel.b.size() - 1]);
}

int sk3DRomModel::read_basic_bone_data( std::ifstream &inf, sk3DRomBone &b)
{
	// Read in the bones joints.
	b.ja.read_binary( inf);
	b.jb.read_binary( inf);

	// Read in the bones joint coordinate system.
	b.jcs.read_binary( inf);

	// Read in th pivot joint.
	inf.read( &(b.pj), 1);
	if( inf.fail()) return -60;

	// Read in the index of the master bone.
	rgWORD n;
	inf.read( (char *)&n, 2);
	b.mid = rgPortLittleToNativeWORD( n);

	// Read in the number of slave bones.
	inf.read( (char *)&n, 2);
	if( inf.fail()) return -65;
	n = rgPortLittleToNativeWORD( n);
	if( short( n) < 0) return -70;
	// Read in all of the slave bones.
	{
		sk3DRomBoneLink *blink;
		const int jstop = n;
		for( int j = 0; j < jstop; j++) {
			// Create a bone link.
			b.link.push_back( sk3DRomBoneLink());
			blink = &( b.link[ b.link.size() - 1]);
				
			inf.read( (char *)&n, 2);
			blink->bid = rgPortLittleToNativeWORD( n);
			inf.read( &(blink->jid), 1);
			if( inf.fail()) return -80;
		}
	}
		
	// Read in the number of controlled verticies
	inf.read( (char *)&n, 2);
	if( inf.fail()) return -90;
	n = rgPortLittleToNativeWORD( n);
	if( short( n) < 0) return -100;
	// Read in all of the controlled verticies
	{
		const int jstop = n;
		for( int j = 0; j < jstop; j++) {
			inf.read( (char *)&n, 2);
			if( inf.fail()) return -110;
			n = rgPortLittleToNativeWORD( n);
			if( short( n) < 0) return -120;
			
			b.vid.push_back( short( n));
		}
	}

	return 0;
}

int sk3DRomModel::read_skel_frames_2_data( std::ifstream &inf)
{
	// Read in the total number of frames.
	rgWORD n;
	inf.read( (char *)&n, 2);
	if( inf.fail()) return -10;
	n = rgPortLittleToNativeWORD( n);
	if( short( n) < 0) return -11;
	const int istop = int( n);	// Total number of skeleton frames

	// Read in the total number of bones per frame.
	inf.read( (char *)&n, 2);
	if( inf.fail()) return -20;
	n = rgPortLittleToNativeWORD( n);
	// FIXME: Should perform check to make sure that number of bones
	//	in each frame is the same as the main skeleton (if it was
	//	loaded already).

	// Read in all of the skeleton frames.
	sk3DRomTag tagSkel;
	sk3DRomTag tagBone;

	sk3DRomSkelFrame *tmp_sf;
	sk3DRomBoneFrame *tmp_bf;
	int skel_final_seek;
	int bone_final_seek;
	int j;
	const int jstop = int( n);	// Total number of bones per frame
	for( int i = 0 ; i < istop; i++) {

		// Read in the frame tag (which MUST be sk3DROM_SFRAME or
		// a later supported tag version).
		if( sk3DRom_read_tag( inf, tagSkel)) return -50;
		if( tagSkel.mode != sk3DROM_SFRAME) return -51;
		if( tagSkel.size < 0) return -52;
		// Determine the end of this frames data.
		skel_final_seek = tagSkel.size + inf.tellg();

		// Create a new skeleton frame.
		skel.sf.push_back( sk3DRomSkelFrame());
		tmp_sf = &(skel.sf[ skel.sf.size() - 1]);

		// Read in the name for the skeleton.
		inf.read( tmp_sf->fname, SK3DROM_SKEL_FRAME_NAME_LEN);
		if( inf.fail()) return -60;
		// The last char in the skeleton frame name MUST be NULL.
		tmp_sf->fname[ SK3DROM_SKEL_FRAME_NAME_LEN - 1] = '\0';

		// Read in all of the bones in this skeleton frame.
		for( j = 0; j < jstop; j++) {

			// Read in the bone frame tag (which MUST be
			// sk3DROM_BONE_FRAME, sk3DROM_BONE_FRAME_2 or a
			// later supported tag version).
			if( sk3DRom_read_tag( inf, tagBone)) return -70;
			switch( tagBone.mode) {
			case sk3DROM_BONE_FRAME:
			case sk3DROM_BONE_FRAME_2: break;
			default: return -71;
			}

			// Determine the end of this bones data.
			if( tagBone.size < 0) return -72;
			bone_final_seek = tagBone.size + inf.tellg();

			// Create a new bone frame.
			tmp_sf->b.push_back( sk3DRomBoneFrame());
			tmp_bf = &( tmp_sf->b[ tmp_sf->b.size() - 1]);

			// Read in the bone's data.
			switch( tagBone.mode) {
			case sk3DROM_BONE_FRAME:
				if( read_bone_frame_data( inf, *tmp_bf))
					return -80;
				break;
			
			case sk3DROM_BONE_FRAME_2:
				if( read_bone_frame_2_data( inf, *tmp_bf))
					return -81;
				break;
			}

			// If this bone frame tag contains more data then this
			// function is currently interested in, then skip the
			// rest of the data.
			if( inf.tellg() < bone_final_seek) {
				inf.seekg( bone_final_seek, std::ios::beg);
				if( inf.fail()) return -500;
			}
		}

		// If this skeleton frame tag contains more data then this
		// function is currently interested in, then skip the rest of
		// the data.
		if( inf.tellg() < skel_final_seek) {
			inf.seekg( skel_final_seek, std::ios::beg);
			if( inf.fail()) return -600;
		}
	}
	
	return 0;
}

int sk3DRomModel::read_bone_frame_data( std::ifstream &inf,
	sk3DRomBoneFrame &bf)
{
	// Read in the bones rotation coordinate system (rcs).
	bf.rcs.read_binary( inf);
	return 0;
}

int sk3DRomModel::read_bone_frame_2_data( std::ifstream &inf,
	sk3DRomBoneFrame &bf)
{
	// Read the bone frame version number
	rgWORD n;
	inf.read( (char *)&n, sizeof( n));

	// Read the original bone frame data.
	if( read_bone_frame_data( inf, bf)) return -1;

	// Read the additional bone frame 2 data.
	rgDWORD d;
	inf.read( (char *)&d, sizeof( d));
	bf.msb = int( d);
	inf.read( (char *)&d, sizeof( d));
	bf.mmb = int( d);
	inf.read( (char *)&d, sizeof( d));
	bf.cms = int( d);
	if( inf.fail()) return -2;
	return 0;
}


int sk3DRomModel::read_skel_animation_data( std::ifstream &inf)
{
	// Read in the total number of sequences
	rgWORD n;
	inf.read( (char *)&n, 2);
	if( inf.fail()) return -10;
	n = rgPortLittleToNativeWORD( n);
	if( short( n) < 0) return -11;
	const int istop = int( n);	// Total number of sequences

	// Read in all of the sequences.
	sk3DRomTag t;
	sk3DRomSequence *tmp_seq;
	int final_seek;
	for( int i = 0 ; i < istop; i++) {

		// Read in the sequence tag (which MUST be sk3DROM_ANI_SEQ or
		// a later supported tag version).
		if( sk3DRom_read_tag( inf, t)) return -50;
		if( t.mode != sk3DROM_ANI_SEQ) return -51;
		if( t.size < 0) return -52;
		// Determine the end of this sequence's data.
		final_seek = t.size + inf.tellg();

		// Create a new sequence
		skel.seq.push_back( sk3DRomSequence());
		tmp_seq = &(skel.seq[i]);

		{ // Read in the name for the sequence.
			char tmp[ SK3DROM_SEQUENCE_NAME_LEN];
			inf.read( tmp, SK3DROM_SEQUENCE_NAME_LEN);
			tmp_seq->set_name( tmp);
		}

		if( read_animation_frame_data( inf, skel.seq.size() - 1))
			return -70;

		// If this sequence tag contains more data then this
		// function is currently interested in, then skip the rest of
		// the data.
		if( inf.tellg() < final_seek) {
			inf.seekg( final_seek, std::ios::beg);
			if( inf.fail()) return -600;
		}
	}

	return 0;
}

int sk3DRomModel::read_animation_frame_data( std::ifstream &inf, const int sid)
{
	// Read in the total number of animation frames.
	const int istop = int( read_word( inf));
	if( istop < 0) return -10;

	// Read in all of the animation frames.
	sk3DRomTag t;
	for( int i = 0 ; i < istop; i++) {

		// Read in the frame tag (which MUST be sk3DROM_ANI_FRAME or
		// a later supported tag version).
		if( sk3DRom_read_tag( inf, t)) return -50;
		if( t.size < 0) return -51;
		// Determine the end of this animation frame's data.
		const int final_seek = t.size + inf.tellg();

		// XXX: sk3DROM_ANI_FRAME is no longer supported. Any data
		//	stored within this tag type will be ignored.
		// FIXME: Eventually, make it an error if a file contains
		//	the old sk3DROM_ANI_FRAME tag.
		//if( t.mode != sk3DROM_ANI_FRAME) return -52;
		if( sk3DROM_ANI_FRAME_2 == t.mode) {
			skel.seq[sid].frame.push_back( sk3DRomAniFrame());
			sk3DRomAniFrame &tmp_frm = skel.seq[sid].frame[
				skel.seq[sid].frame.size() - 1];

			switch( read_word( inf)) {
			case SK3DROM_BONE_FRAME_VERSION_A:
				tmp_frm.set_fid( read_dword( inf));
				tmp_frm.set_time( read_word( inf));
				tmp_frm.set_code( read_word( inf));
				tmp_frm.data[0] = read_dword( inf);
				tmp_frm.data[1] = read_dword( inf);
				tmp_frm.data[2] = read_dword( inf);
				tmp_frm.data[3] = read_dword( inf);
				tmp_frm.data[4] = read_dword( inf);
				tmp_frm.data[5] = read_dword( inf);
				if( inf.fail()) return -60;
				break;
			}
		}

		// If this animation frame tag contains more data then this
		// function is currently interested in, then skip the rest of
		// the data.
		if( inf.tellg() < final_seek) {
			inf.seekg( final_seek, std::ios::beg);
			if( inf.fail()) return -600;
		}
	}

	return 0;
}

int sk3DRomModel::read_model_texture_data( std::istream &inf)
{
	const short total_textures = short( read_word( inf));
	if( total_textures < 0) return -11;

	for( int i = 0 ; i < total_textures; i++) {

		// Read in the sequence tag (which MUST be sk3DROM_TEXTURE_2 or
		// a later supported tag version).
		sk3DRomTag t;
		if( sk3DRom_read_tag( inf, t)) return -50;
		if( t.size < 0) return -54;
		// Determine the end of this textures's data.
		const int final_seek = t.size + inf.tellg();

		if( sk3DROM_TEXTURE_2 == t.mode) {
			const int err = read_texture_2_data( inf);
			if( err) return err;
		
		} else if( sk3DROM_TEXTURE_3 == t.mode) {
			const int err = read_texture_3_data( inf);
			if( err) return err;

		} else return -53;

		// If this texture tag contains more data then this function
		// is currently interested in, then skip the rest of the data.
		if( inf.tellg() < final_seek) {
			inf.seekg( final_seek, std::ios::beg);
			if( inf.fail()) return -600;
		}
	}

	return 0;
}

int sk3DRomModel::read_texture_2_data( std::istream &inf)
{
	sk3DRomTexture tex;

	const short file_name_len = short( read_word( inf));
	if( file_name_len < 0) return -540;
	if( tex.read_file_name( inf, file_name_len)) return -550;

	tex.set_flags( read_dword( inf));

	if( inf.fail()) return -560;
	texture.push_back( tex);
	return 0;
}

int sk3DRomModel::read_texture_3_data( std::istream &inf)
{
	sk3DRomTexture tex;

	const int version = read_dword( inf);
	if(	(SK3DROM_TEXTURE_VERSION_A == version) ||
		(SK3DROM_TEXTURE_VERSION_B == version)) {
		const short file_name_len = short( read_word( inf));
		if( file_name_len < 0) return -540;
		if( tex.read_file_name( inf, file_name_len)) return -541;

		const short tex_name_len = short( read_word( inf));
		if( tex_name_len < 0) return -545;
		if( tex.read_tex_name( inf, tex_name_len)) return -546;

		tex.set_flags( read_dword( inf));
		if( SK3DROM_TEXTURE_VERSION_B == version) {
			inf.read( (char *)tex.m_colorkey, 4);
		}

	} else return -600;

	texture.push_back( tex);
	if( inf.fail()) return -610;
	return 0;
}

int sk3DRomModel::read_face_data( std::ifstream &inf)
{
	short count;
	inf.read( (char *)&count, 2);
	if( inf.fail()) return -1;

	#ifdef FACE_DEBUG
	std::cerr << "FACE\n";
	std::cerr << "\tcount = " << count << "\n";
	#endif

	{ // Read in all of the face data
		sk3DRomFace temp;
		while( count--) {
			// Read 3 vertex id's (2 bytes per vertex)
			inf.read( (char *)&(temp.vid), 6);
			if( inf.fail()) return -2;

			// Read 3 vertex normals
			(temp.n[0]).read_binary( inf);
			(temp.n[1]).read_binary( inf);
			(temp.n[2]).read_binary( inf);

			// Read 3 vertex lights (RGBA), 4 bytes per light
			inf.read( (char *)(temp.li), 12);
			if( inf.fail()) return -8;

			// Read texture id (2 bytes)
			inf.read( (char *)&(temp.tid), 2);
			if( inf.fail()) return -9;

			// Read 3 mapping coords (8 bytes per map coord)
			inf.read( (char *)(temp.map), 24);
			if( inf.fail()) return -10;

			f.push_back( temp);
		}
	}
	
	return 0;
}

int sk3DRomModel::read_face_2_data( std::ifstream &inf)
{
	short count = read_word( inf);
	#ifdef FACE_DEBUG
	std::cerr << "FACE\n";
	std::cerr << "\tcount = " << count << "\n";
	#endif
	if( count < 0) return -100;

	// XXX: Load faces by version.
	const int version = read_dword( inf);
	while( count--) {
		sk3DRomFace temp;
		switch( version) {
		case SK3DROM_FACE_VERSION_A:
			if( read_version_A_face( inf, temp)) return -2;
			break;

		case SK3DROM_FACE_VERSION_B:
			if( read_version_B_face( inf, temp)) return -3;
			break;

		default: return -200;	// XXX: Unsupported face version
		}
		f.push_back( temp);
	}
	return 0;
}

int sk3DRomModel::read_version_A_face( std::ifstream &inf, sk3DRomFace &p_face)
{
	// Read 3 vertex id's (2 bytes per vertex).
	inf.read( (char *)&(p_face.vid), 6);
	if( inf.fail()) return -2;

	// Read 3 vertex normals.
	(p_face.n[0]).read_binary( inf);
	(p_face.n[1]).read_binary( inf);
	(p_face.n[2]).read_binary( inf);

	// Read 3 vertex lights (RGBA), 4 bytes per light.
	inf.read( (char *)(p_face.li), 12);
	if( inf.fail()) return -8;

	p_face.tid = read_word( inf);	// Read texture id (2 bytes)
	if( p_face.tid < -1) p_face.tid = -1;

	// Read 3 mapping coords (8 bytes per map coord).
	inf.read( (char *)(p_face.map), 24);
	if( inf.fail()) return -10;

	p_face.flags = read_dword( inf);
	return 0;
}

int sk3DRomModel::read_version_B_face( std::ifstream &inf, sk3DRomFace &p_face)
{
	if( read_version_A_face( inf, p_face)) return -1;
	p_face.gid = read_dword( inf);
	return 0;
}

bool sk3DRomModel::load_texture_from_paths( sk3DRomTexture &t,
	const char **paths, int count) const
{
	const int len = 300;
	char file_buf[ len];
	file_buf[ len - 1] = '\0';
	t.filename( file_buf, len);
	while( (count--) > 0) {
		if( (*paths)) {
			std::string s( (*paths));
			if( !rg_slash_terminates_unix_path( s.c_str()))
				s += '/';
			// FIXME: Use "append" instead of "+" operator since
			//	it "should" be more portable across C++
			//	implementations.
			//#ifdef WINDOWS_OS
			s.append( file_buf);

			if( RGERR_SUCCESS == t.load( s.c_str())) return true;
			//if( RGERR_SUCCESS ==
			//	(t.load( (s.append( file_buf)).c_str())))
			//	return true;
			//#else
			//if( RGERR_SUCCESS == (t.load( (s + file_buf).
			//c_str())))
			//	return true;
			//#endif
		}
		paths++;
	}
	return false;
}

int sk3DRomModel::load_textures( const char **paths, const int count)
{
	if( texture.empty()) return 0;

	int result = 0;
	for( int i = 0; i < int( texture.size()); i++) {
		// Try all given paths
		if( paths) {
			if( !load_texture_from_paths( texture[i], paths,
				count)) result = -1;

		// Load the texture from its absolute location.
		} else if( texture[i].load() != RGERR_SUCCESS) result = -1;
	}

	return result;
}

int sk3DRomModel::read_group_data( std::ifstream &inf)
{
	// Read the total number of groups.
	rgDWORD d;
	inf.read( (char *)&d, sizeof( d));
	if( inf.fail()) return -1;

	// Read in all of the groups now.
	int count = int( d);
	if( count < 0) return -2;
	while( count--) {

		{ // Read in a single group
			sk3DRomTag t;
			if( sk3DRom_read_tag( inf, t)) return -10;
			if( t.mode != sk3DROM_GROUP) return -11;
			if( t.size < 0) return -12;
			const int final_seek = t.size + inf.tellg();

			sk3DRomGroup grp;
			inf.read( grp.gname, SK3DROM_GROUP_NAME_MAX_LEN);
			inf.read( (char *)(grp.gcolor), 4);
			inf.read( (char *)&d, sizeof( d));
			grp.flags = rgPortLittleToNativeDWORD( d);
			if( inf.fail()) return -20;

			group.push_back( grp);

			if( inf.tellg() < final_seek) {
				inf.seekg( final_seek, std::ios::beg);
				if( inf.fail()) return -600;
			}
		}
	}

	return 0;
}

int sk3DRomModel::read_pages_data( std::ifstream &inf)
{
	// Read in the total number of pages.
	int count = int( read_dword( inf));
	if( count < 0) return -2;
	while( count--) { // Read in all of the pages now.

		sk3DRomTag t;
		if( sk3DRom_read_tag( inf, t)) return -10;
		if( t.mode != sk3DROM_PAGE) return -11;
		if( t.size < 0) return -12;
		const int final_seek = t.size + inf.tellg();

		m_page.push_back( sk3DRomTexturePage());
		sk3DRomTexturePage &p = m_page[ page_count() - 1];
		switch( read_dword( inf)) {
		case SK3DROM_PAGE_VERSION_A:
			if( p.read_versionA_page( inf)) return -20;
			break;
		}

		if( inf.tellg() < final_seek) {
			inf.seekg( final_seek, std::ios::beg);
			if( inf.fail()) return -600;
		}
	}

	if( inf.fail()) return -1000;
	return 0;
}

// XXX: Gets the "type" of brush, and loads as appropriate (unknown brushes
//	are ignored).
inline int hlp_load_brush( std::ifstream &inf, rgCSGWorld &csg,
	sk3DRomModel *mdl)
{
	rgCSGBrush *b;
	bool tmp_ignore_unknown_brush = false;
	switch( read_dword( inf)) {
	// XXX: ADDING NEW BRUSHES
	//	Loading of brushes is done here. Add a new "case" statment for
	//	your new brush.
	case RG3DROM_CSG_CUBE: {
			sk3DRomCSGCube *tmp = new sk3DRomCSGCube();
			tmp->m_model = mdl;
			csg.add_brush( (b = tmp)); break;
		}
	default:
		tmp_ignore_unknown_brush = true;
		b = 0;	// XXX: Required since g++ gives warning on -O3
			//	compiler optimization flag.
		break;
	}
	if( !tmp_ignore_unknown_brush) return (b->load_data( inf));
	return 0;
}

int sk3DRomModel::read_csg_data( std::ifstream &inf)
{
	const int istop = read_dword( inf);
	for( int i = 0; i < istop; i++) {
		sk3DRomTag t;
		if( sk3DRom_read_tag( inf, t)) return -10;
		if( t.mode != sk3DROM_CSG_BRUSH) return -11;
		if( t.size < 0) return -12;
		const int final_seek = t.size + inf.tellg();

		if( hlp_load_brush( inf, *(m_csg), this)) return -20;

		if( inf.tellg() < final_seek) {
			inf.seekg( final_seek, std::ios::beg);
			if( inf.fail()) return -600;
		}
	}
	if( inf.fail()) return -1000;
	return 0;
}
