/*
Copyright (C) 2004-2005 Amour Hassan

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#include <rgPort.h>
#include <iostream>
#include <fstream>
#include <string.h>
#include <rgOpenGL.h>
#include <rgBSP.h>
#include <rgBSP_template.cpp>
#include "rg3DRom.h"
#include "rg3DRom_vertex_array.h"

int hlp_get_face_bone( sk3DRomFace &f, sk3DRomSkel &s);

void hlp_seperate_BSP_into_BSP_bone_lists( sk3DRomModel &mdl,
        std::vector <rgBSPFace <rgBSP_NTTC_ExtData> > &bspFace,
        std::vector <int> &bone_id);

int sk3DRomModel::get_render_format()
{
	return render_format;
}

rgVector sk3DRomModel::get_tv_vertex( const int i) const
{
	if( !tv) return rgVector();
	switch( render_format) {
	// FIXME: "Render-format" is now deprecated. This is because
	//	sk3DRomModel does not have to directly provide facilities for
	//	optimal rendering.
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		{
			rgFloat *tv_ptr = tv + (i << 4);
			return rgVector( tv_ptr[0], tv_ptr[1], tv_ptr[2]);
		}
		break;
	}
	return rgVector();
}

/* FIXME: "set_vertex" has a bad running time due to rebinding. It is more
	efficient to rebind multiple verticies all at once rather than one
	at a time.
void sk3DRomModel::set_vertex( const int i, const rgVector p_v)
{
	v[i].x = p_v.x;
	v[i].y = p_v.y;
	v[i].z = p_v.z;

	// Rebind the vertex if it is owned/controlled by a bone.
	if( v[i].bid < 0) return;
	
}
*/

void sk3DRomModel::set_tv_vertex( const int i, const rgVector p_v)
{
	if( !tv) transform_skel_frame();
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		{
			rgFloat *tv_ptr = tv + (i << 4);
			tv_ptr[0] = p_v.x;
			tv_ptr[1] = p_v.y;
			tv_ptr[2] = p_v.z;
		}
		break;
	}
}

void sk3DRomModel::set_vertex_from_tv( const int *vlist, int count)
{
	set_vertex_from_tv( vlist, count, skel.skout);
}

void sk3DRomModel::set_vertex_from_tv( const int *vlist, int count,
	const int sid)
{
	if( sid < 0) set_vertex_from_tv( vlist, count, skel.skout);
	else if( sid >= int( skel.sf.size())) return;
	set_vertex_from_tv( vlist, count, skel.sf[sid]);
}

void sk3DRomModel::set_vertex_from_tv( const int *vlist, int count,
	const sk3DRomSkelFrame &sf)
{
	// NOTE: The transformation array "tv" stores ALL verticies relative
	//	to the standard coordinate system. Therefore, all verticies
	//	set to their "tv" equivalent must be "rebound" to their
	//	respective bones.
	if( !vlist) return;
	if( count <= 0) return;
	rgCoordSystem *cs = 0;
	char *ready = 0;
	if( !(skel.b.empty())){
		cs = new rgCoordSystem[ skel.b.size()];
		ready = new char[ skel.b.size()];
		memset( ready, 0, skel.b.size());
	}

	// Set the vertices and precalculate the jcs for bones that require
	// it. This way, multiple vertices which share a bone will only cause
	// a single processing of the bone jcs (which is a time consuming
	// process).
	int i, bid;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		for( i = 0; i < count; i++) {
			v[ vlist[i]].x = tv[ vlist[i] << 4];
			v[ vlist[i]].y = tv[ (vlist[i] << 4) + 1];
			v[ vlist[i]].z = tv[ (vlist[i] << 4) + 2];
			bid = v[ vlist[i]].bid();
			if( (bid >= 0) && (!(ready[bid]))) {
				ready[bid] = 1;
				cs[bid] = skel.get_full_bone_rcs( bid, sf);
			}
		}
		break;
	}
	if( ready) delete [] ready;

	// Run through all verticies again and this time bind them (if needed).
	while( count--) {
		if( v[ *vlist].bid() >= 0) {
			v[ *vlist] = cs[ v[ *vlist].bid()].change_point_csys(
				v[ *vlist]);
		}
		vlist++;
        }
	if( cs) delete [] cs;
}

inline void hlp_render_bone_frame( const sk3DRomSkel &s,
	const sk3DRomSkelFrame &sf, rgCoordSystem *mcs, const int bid)
{
	rgCoordSystem cs( sf.b[bid].rcs);
	if( mcs) cs = mcs->map_csys( cs);
	const sk3DRomBone &b = s.b[bid];
	glVertex( cs.map_point( b.ja));
	glVertex( cs.map_point( b.jb));
}

void hlp_render_slave_bone_frame( const sk3DRomSkel &s,
	const sk3DRomSkelFrame &sf, rgCoordSystem *mcs, const int bid)
{
	rgCoordSystem cs( sf.b[bid].rcs);
	if( mcs) cs = mcs->map_csys( cs);
	const std::vector <sk3DRomBoneLink> &link = s.b[bid].link;
	const int istop = int( link.size());
	for( int i = 0; i < istop; i++) {
		hlp_render_bone_frame( s, sf, &cs, link[i].bid);
		if( !(s.b[ link[i].bid].link.empty()))
			hlp_render_slave_bone_frame( s, sf, &cs, link[i].bid);
	}
}

// Description: Renders a skeleton frame by rendering all of its bones. Bones
//	are rendered starting with root bones, and then followed by slave
//	bones. This is efficient since the alternative would require every
//	slave bone to recurse down to its primary master bone whenever the
//	slave needed to determine its absolute coordinate position.
void sk3DRomModel::render_skel_frame()
{
	render_skel_frame( -1);
}

// TODO: Implement efficient root bone rendering order (with mounts).
void sk3DRomModel::render_skel_frame( const int sid)
{
	const sk3DRomSkelFrame *s = &(skel.skout);
	if( sid >= 0) s = &(skel.sf[ sid]);
	/*
	rgCoordSystem *cs = new rgCoordSystem[ skel.b.size()];
	const int istop = int( skel.rb.size());
	for( int i = 0; i < istop; i++) {
		hlp_render_bone_frame( skel, *s, 0, skel.rb[i]);
		hlp_render_slave_bone_frame( skel, *s, 0, skel.rb[i]);
	}
	delete [] cs;
	*/

	const int istop = int( skel.b.size());
	for( int i = 0; i < istop; i++) {
		{
			const rgCoordSystem cs(
				skel.get_full_bone_rcs( i, *s));
			const sk3DRomBone &b = skel.b[i];
			glVertex( cs.map_point( b.ja));
			glVertex( cs.map_point( b.jb));
		}
	}
}

// Description: Renders a single bone from a given skeleton frame. If the
//	entire skeleton is to be rendered, then use "render_skel_frame"
//	instead. This is because "render_bone" has a high overhead for each
//	bone rendered (because it must traverse through master bones to
//	determine the bone's "rcs").
void sk3DRomModel::render_bone( const sk3DRomSkelFrame &sf, const int bid)
{
	if( (bid < 0) || (bid >= int( sf.b.size()))) return;
	//rgCoordSystem cs( skel.get_full_bone_rcs( skel.b[bid].mid, sf));
	rgCoordSystem cs;
	skel.get_full_bone_rcs( bid, sf, cs);
	hlp_render_bone_frame( skel, sf, &cs, bid);
}

void sk3DRomModel::offset_tv_verticies( const int *vlist, int count,
	const rgVector p_v)
{
	if( !vlist) return;
	if( count <= 0) return;
	if( !tv) transform_skel_frame();
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		{
			while( count--) {
				tv[ ((*vlist) << 4)] += p_v.x;
				tv[ ((*vlist) << 4) + 1] += p_v.y;
				tv[ ((*vlist) << 4) + 2] += p_v.z;
				vlist++;
			}
		}
		break;
	}
}

// XXX: Fix spelling mistake "verticies".
void sk3DRomModel::get_tv_face_verticies( const int i, rgVector *p_v)
{
	if( !tv) { transform_skel_frame(); if( !tv) return;}
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		{
			rgFloat *tv_ptr = tv + (int( f[i].vid[0]) << 4);
			p_v[0] = rgVector( tv_ptr[0], tv_ptr[1], tv_ptr[2]);

			tv_ptr = tv + (int( f[i].vid[1]) << 4);
			p_v[1] = rgVector( tv_ptr[0], tv_ptr[1], tv_ptr[2]);

			tv_ptr = tv + (int( f[i].vid[2]) << 4);
			p_v[2] = rgVector( tv_ptr[0], tv_ptr[1], tv_ptr[2]);
			return;
		}
		break;
	}
}

void sk3DRomModel::get_tv_bounding_box( float &left, float &right,
	float &bottom,float &top, float &front, float &back)
{
	if( !tv) transform_skel_frame();
	//if( !tv) {
	//	left = right = bottom = top = front = back = 0.0f;
	//	return;
	//}

	left = 16777815.0f;
	right = -16777815.0f;
	bottom = 16777815.0f;
	top = -16777815.0f;
	front = -16777815.0f;
	back = 16777815.0f;

	const int istop = int( v.size());
	int i;
	rgFloat *tv_ptr = tv;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		for( i = 0; i < istop; i++) {
			if( tv_ptr[0] < left) left = tv_ptr[0];		// X
			if( tv_ptr[0] > right) right = tv_ptr[0];	// X

			if( tv_ptr[1] < bottom) bottom = tv_ptr[1];	// Y
			if( tv_ptr[1] > top) top = tv_ptr[1];		// Y

			if( tv_ptr[2] < back) back = tv_ptr[2];		// Z
			if( tv_ptr[2] > front) front = tv_ptr[2];	// Z
			
			tv_ptr += 16;
		}
		return;
	}
}

void sk3DRomModel::reveal_all_vertices()
{
	const int istop = int( v.size());
	for( int i = 0; i < istop; i++) {
		if( (v[i].gid() < 0) || !(group[ v[i].gid()].hidden()))
			v[i].flags &= ~SK3DROM_VERT_HIDDEN_BIT;
	}
}

#define TMP_IS_HIDDEN(v) (((v).flags) & SK3DROM_VERT_HIDDEN_BIT)
void sk3DRomModel::render_verts()
{
	if( !tv) { transform_skel_frame(); if( !tv) return;}

	const int istop = int( v.size());
	int i;
	rgFloat *tv_ptr = tv;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		for( i = 0; i < istop; i++) {
			if( !TMP_IS_HIDDEN( v[i])) glVertex3fv( tv_ptr);
			tv_ptr += 16;
		}
		return;
	}
}

void sk3DRomModel::render_verts( const int *vlist, const int count)
{
	if( count < 0) return;
	if( !tv) { transform_skel_frame(); if( !tv) return;}

	int i;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		for( i = 0; i < count; i++) {
			if( !TMP_IS_HIDDEN( v[ vlist[i]]))
				glVertex3fv( tv + (vlist[i] << 4));
		}
		return;
	}
}

void sk3DRomModel::render_group_verts()
{
	if( !tv) { transform_skel_frame(); if( !tv) return;}
	const int istop = int( v.size());
	int i;
	rgFloat *tv_ptr = tv;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		for( i = 0; i < istop; i++) {
			if( !TMP_IS_HIDDEN( v[i])) {
				const int gid = v[i].gid();
				if( gid >= 0) glColor3ubv( group[gid].gcolor);
				else glColor3ub( 0, 0, 0);
				glVertex3fv( tv_ptr);
			}
			tv_ptr += 16;
		}
		return;
	}
}


void sk3DRomModel::render_group_verts( const int *vlist, const int count)
{
	if( count < 0) return;
	if( !tv) { transform_skel_frame(); if( !tv) return;}

	int i;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		for( i = 0; i < count; i++) {
			if( !TMP_IS_HIDDEN( v[ vlist[i]])) {
				const int gid = v[ vlist[i]].gid();
				if( gid >= 0) glColor3ubv( group[gid].gcolor);
				else glColor3ub( 0, 0, 0);
				glVertex3fv( tv + (vlist[i] << 4));
			}
		}
		return;
	}
}

#undef TMP_IS_HIDDEN

void sk3DRomModel::render_single_face( bool use_normals, bool use_texture,
	const int i)
{
	render_single_face( use_normals, use_texture, i, 0, 0, 0);
}

void sk3DRomModel::render_single_face( bool use_normals, bool use_texture,
	const int i, float *vc1, float *vc2, float *vc3)
{
	if( !tv) { transform_skel_frame(); if( !tv) return;}

	if( !use_normals && !use_texture) {
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			if( vc1) glColor3fv( vc1);
			glVertex3fv( tv + (int( f[i].vid[0]) << 4));

			if( vc2) glColor3fv( vc2);
			glVertex3fv( tv + (int( f[i].vid[1]) << 4));

			if( vc3) glColor3fv( vc3);
			glVertex3fv( tv + (int( f[i].vid[2]) << 4));
			return;
		}

	} else if( use_normals && !use_texture) {
		rgFloat *tv_ptr;
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			tv_ptr = tv + (int( f[i].vid[0]) << 4);
			if( vc1) glColor3fv( vc1);
			glNormal3fv( tv_ptr + 4);
			glVertex3fv( tv_ptr);

			tv_ptr = tv + (int( f[i].vid[1]) << 4);
			if( vc2) glColor3fv( vc2);
			glNormal3fv( tv_ptr + 4);
			glVertex3fv( tv_ptr);

			tv_ptr = tv + (int( f[i].vid[2]) << 4);
			if( vc3) glColor3fv( vc3);
			glNormal3fv( tv_ptr + 4);
			glVertex3fv( tv_ptr);
			return;
		}

	} else if( !use_normals && use_texture) {
		rgFloat *tv_ptr;
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			tv_ptr = tv + (int( f[i].vid[0]) << 4);
			if( vc1) glColor3fv( vc1);
			glTexCoord2fv( tv_ptr + 8);
			glVertex3fv( tv_ptr);

			tv_ptr = tv + (int( f[i].vid[1]) << 4);
			if( vc2) glColor3fv( vc2);
			glTexCoord2fv( tv_ptr + 8);
			glVertex3fv( tv_ptr);

			tv_ptr = tv + (int( f[i].vid[2]) << 4);
			if( vc3) glColor3fv( vc3);
			glTexCoord2fv( tv_ptr + 8);
			glVertex3fv( tv_ptr);
			return;
		}

	} else {
		rgFloat *tv_ptr;
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			tv_ptr = tv + (int( f[i].vid[0]) << 4);
			if( vc1) glColor3fv( vc1);
			glTexCoord2fv( tv_ptr + 8);
			glNormal3fv( tv_ptr + 4);
			glVertex3fv( tv_ptr);

			tv_ptr = tv + (int( f[i].vid[1]) << 4);
			if( vc2) glColor3fv( vc2);
			glTexCoord2fv( tv_ptr + 8);
			glNormal3fv( tv_ptr + 4);
			glVertex3fv( tv_ptr);

			tv_ptr = tv + (int( f[i].vid[2]) << 4);
			if( vc3) glColor3fv( vc3);
			glTexCoord2fv( tv_ptr + 8);
			glNormal3fv( tv_ptr + 4);
			glVertex3fv( tv_ptr);
			return;
		}

	}
}


// NOTE: For best performance, faces should be grouped together by their
//	texture id's (this minimizes texture swaping).
#define HLP_BIND_TEXTURE_IF_NEEDED( fid) \
	if( (f[(fid)].tid) != act_tex) { \
		glEnd(); \
		if( f[(fid)].tid < 0) glBindTexture( GL_TEXTURE_2D, 0); \
		else texture[ f[(fid)].tid].bind(); \
		glBegin( GL_TRIANGLES); \
		act_tex = f[(fid)].tid; \
	}

#define TMP_BSP_FACE(i) (f[i].is_bsp())

#define TMP_USE_FACE_LIST (flist && (count > 0))

//////////////////////////////////////////////////////////////////////////////
// Core rendering function for V faces
//////////////////////////////////////////////////////////////////////////////
#define TMP_RENDER_CODE(i) \
	if( draw_test_fn( *this, i)) { \
		if( use_group_color) set_face_group_color3( i); \
		glVertex3fv( tv + (int( f[i].vid[0]) << 4)); \
		glVertex3fv( tv + (int( f[i].vid[1]) << 4)); \
		glVertex3fv( tv + (int( f[i].vid[2]) << 4));}
void sk3DRomModel::hlp_render_faces_V( const bool use_group_color,
	const bool block_bsp_face,
	const int *flist, const int count,
	bool (*draw_test_fn)( const sk3DRomModel &, const int))
	const
{
	//std::vector <sk3DRomFace> &f = mdl.f;
	//rgFloat *tv = mdl.tv;
	const int istop = int( f.size());
	if( block_bsp_face) {
		if( TMP_USE_FACE_LIST) {
			for( int i = 0; i < count; i++) {
				if( !TMP_BSP_FACE( flist[i]))
					TMP_RENDER_CODE( flist[i]);
			}

		} else {
			for( int i = 0; i < istop; i++) {
				if( !TMP_BSP_FACE( i)) TMP_RENDER_CODE( i);
			}
		}

	} else {
		if( TMP_USE_FACE_LIST) {
			for( int i = 0; i < count; i++)
				TMP_RENDER_CODE( flist[i]);
		} else {
			for( int i = 0; i < istop; i++) TMP_RENDER_CODE( i);
		}
	}
}
#undef TMP_RENDER_CODE


//////////////////////////////////////////////////////////////////////////////
// Core rendering function for VN faces
//////////////////////////////////////////////////////////////////////////////
#define TMP_RENDER_CODE(i) \
	if( draw_test_fn( *this, i)) { \
		if( use_group_color) set_face_group_color3( i); \
		glNormal( f[i].n[0]); \
		glVertex3fv( tv + (int( f[i].vid[0]) << 4)); \
		glNormal( f[i].n[1]); \
		glVertex3fv( tv + (int( f[i].vid[1]) << 4)); \
		glNormal( f[i].n[2]); \
		glVertex3fv( tv + (int( f[i].vid[2]) << 4));}
void sk3DRomModel::hlp_render_faces_VN( const bool use_group_color,
	const bool block_bsp_face,
	const int *flist, const int count,
	bool (*draw_test_fn)( const sk3DRomModel &, const int))
	const
{
	//std::vector <sk3DRomFace> &f = mdl.f;
	//const rgFloat *tv = mdl.tv;
	const int istop = int( f.size());
	if( block_bsp_face) {
		if( TMP_USE_FACE_LIST) {
			for( int i = 0; i < count; i++) {
				if( !TMP_BSP_FACE( flist[i]))
					TMP_RENDER_CODE( flist[i]);
			}

		} else {
			for( int i = 0; i < istop; i++) {
				if( !TMP_BSP_FACE( i)) TMP_RENDER_CODE( i);
			}
		}

	} else {
		if( TMP_USE_FACE_LIST) {
			for( int i = 0; i < count; i++)
				TMP_RENDER_CODE( flist[i]);
		} else {
			for( int i = 0; i < istop; i++) TMP_RENDER_CODE( i);
		}
	}
}
#undef TMP_RENDER_CODE

//////////////////////////////////////////////////////////////////////////////
// Core rendering function for VT faces
//////////////////////////////////////////////////////////////////////////////
#define TMP_RENDER_CODE(i) \
	if( draw_test_fn( *this, i)) { \
		if( use_group_color) set_face_group_color3( i); \
		tv_ptr = tv + (int( f[i].vid[0]) << 4);\
		glTexCoord2f( f[i].map[0], f[i].map[1]);\
		glVertex3fv( tv_ptr);\
		tv_ptr = tv + (int( f[i].vid[1]) << 4);\
		glTexCoord2f( f[i].map[2], f[i].map[3]);\
		glVertex3fv( tv_ptr);\
		tv_ptr = tv + (int( f[i].vid[2]) << 4);\
		glTexCoord2f( f[i].map[4], f[i].map[5]);\
		glVertex3fv( tv_ptr);}
void sk3DRomModel::hlp_render_faces_VT( const bool use_group_color,
	const bool block_bsp_face,
	const int *flist, const int count,
	bool (*draw_test_fn)( const sk3DRomModel &, const int))
	const
{
	//std::vector <sk3DRomFace> &f = mdl.f;
	//const rgFloat *tv = mdl.tv;
	const rgFloat *tv_ptr;
	int act_tex = -1;
	const int istop = int( f.size());
	if( block_bsp_face) {
		if( TMP_USE_FACE_LIST) {
			for( int i = 0; i < count; i++) {
				if( !TMP_BSP_FACE( flist[i])) {
					HLP_BIND_TEXTURE_IF_NEEDED( flist[i]);
					TMP_RENDER_CODE( flist[i]);
				}
			}
		} else {
			for( int i = 0; i < istop; i++) {
				if( !TMP_BSP_FACE( i)) {
					HLP_BIND_TEXTURE_IF_NEEDED( i);
					TMP_RENDER_CODE( i);
				}
			}
		}

	} else {
		if( TMP_USE_FACE_LIST) {
			for( int i = 0; i < count; i++) {
				HLP_BIND_TEXTURE_IF_NEEDED( flist[i]);
				TMP_RENDER_CODE( flist[i]);
			}
		} else {
			for( int i = 0; i < istop; i++) {
				HLP_BIND_TEXTURE_IF_NEEDED( i);
				TMP_RENDER_CODE( i);
			}
		}
	}
}
#undef TMP_RENDER_CODE

//////////////////////////////////////////////////////////////////////////////
// Core rendering function for VNT faces
//////////////////////////////////////////////////////////////////////////////
#define TMP_RENDER_CODE(i) \
	if( draw_test_fn( *this, i)) { \
		if( use_group_color) set_face_group_color3( i); \
		tv_ptr = tv + (int( f[i].vid[0]) << 4);\
		glTexCoord2f( f[i].map[0], f[i].map[1]);\
		glNormal( f[i].n[0]);\
		glVertex3fv( tv_ptr);\
		tv_ptr = tv + (int( f[i].vid[1]) << 4);\
		glTexCoord2f( f[i].map[2], f[i].map[3]);\
		glNormal( f[i].n[1]);\
		glVertex3fv( tv_ptr);\
		tv_ptr = tv + (int( f[i].vid[2]) << 4);\
		glTexCoord2f( f[i].map[4], f[i].map[5]);\
		glNormal( f[i].n[2]);\
		glVertex3fv( tv_ptr);}
void sk3DRomModel::hlp_render_faces_VNT( const bool use_group_color,
	const bool block_bsp_face,
	const int *flist, const int count,
	bool (*draw_test_fn)( const sk3DRomModel &, const int))
	const
{
	//std::vector <sk3DRomFace> &f = mdl.f;
	//const rgFloat *tv = mdl.tv;
	const rgFloat *tv_ptr;
	int act_tex = -1;
	const int istop = int( f.size());
	if( block_bsp_face) {
		if( TMP_USE_FACE_LIST) {
			for( int i = 0; i < count; i++) {
				if( !TMP_BSP_FACE( flist[i])) {
					HLP_BIND_TEXTURE_IF_NEEDED( flist[i]);
					TMP_RENDER_CODE( flist[i]);
				}
			}
		} else {
			for( int i = 0; i < istop; i++) {
				if( !TMP_BSP_FACE( i)) {
					HLP_BIND_TEXTURE_IF_NEEDED( i);
					TMP_RENDER_CODE( i);
				}
			}
		}
	} else {
		if( TMP_USE_FACE_LIST) {
			for( int i = 0; i < count; i++) {
				HLP_BIND_TEXTURE_IF_NEEDED( flist[i]);
				TMP_RENDER_CODE( flist[i]);
			}
		} else {
			for( int i = 0; i < istop; i++) {
				HLP_BIND_TEXTURE_IF_NEEDED( i);
				TMP_RENDER_CODE( i);
			}
		}
	}
}
#undef TMP_RENDER_CODE
#undef TMP_HIDDEN_TEST
#undef TMP_USE_FACE_LIST
#undef TMP_BSP_FACE

void sk3DRomModel::render_faces( bool use_normals, bool use_texture)
{
	render_faces( use_normals, use_texture, 0, 0, false);
}

void sk3DRomModel::render_faces( bool use_normals, bool use_texture,
	const int *flist, const int count)
{
	render_faces( use_normals, use_texture, flist, count, false);
}

void sk3DRomModel::render_faces( bool use_normals, bool use_texture,
	const int *flist, const int count, bool force_no_bsp)
{
	render_faces( use_normals, use_texture, flist, count, force_no_bsp,
		false);
}

void sk3DRomModel::render_faces( bool use_normals, bool use_texture,
	const int *flist, const int count, bool force_no_bsp,
	bool use_group_color)
{
	render_faces( use_normals, use_texture, flist, count, force_no_bsp,
		use_group_color, 0);
}

void sk3DRomModel::render_faces( bool use_normals, bool use_texture,
	const int *flist, const int count, bool force_no_bsp,
	bool use_group_color, bool (*draw_test_fn)( const sk3DRomModel &,
	const int i))
{
	if( count < 0) return;
	if( !tv) { transform_skel_frame(); if( !tv) return;}

	if( !draw_test_fn) draw_test_fn = sk3DRomModel_face_draw_if_visible;
	if( use_group_color) {
		float tmp[4];
		glGetFloatv( GL_CURRENT_COLOR, tmp);
		m_default_groupless_color = rgVector( tmp[0], tmp[1], tmp[2]);
	}

	glBegin( GL_TRIANGLES);
	const bool block_bsp_face = !(bsp_tree.empty()) && !force_no_bsp;
	if( !use_normals && !use_texture) {
	// Render faces only (no normals or textures)
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			hlp_render_faces_V( use_group_color, block_bsp_face,
				flist, count, draw_test_fn);
			break;
		}

	} else if( use_normals && !use_texture) {
	// Render faces and normals (no textures)
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			hlp_render_faces_VN( use_group_color, block_bsp_face,
				flist, count, draw_test_fn);
			break;
		}

	} else if( !use_normals && use_texture) {
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			hlp_render_faces_VT( use_group_color, block_bsp_face,
				flist, count, draw_test_fn);
			break;
		}

	} else {
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			hlp_render_faces_VNT( use_group_color, block_bsp_face,
				flist, count, draw_test_fn);
			break;
		}

	}
	glEnd();
}

bool sk3DRomModel_face_draw_if_visible( const sk3DRomModel &mdl, const int fid)
{
	return !(mdl.face_hidden( fid));
}

bool sk3DRomModel_face_draw_if_unmapped( const sk3DRomModel &mdl,
	const int fid)
{
	if( mdl.f[ fid].mapped()) return false;
	return true;
}

bool sk3DRomModel_face_draw_if_mapped( const sk3DRomModel &mdl, const int fid)
{
	if( mdl.f[ fid].mapped()) return false;
	return true;
}

/*
void sk3DRomModel::render_faces( bool use_normals, bool use_texture)
{
	if( !tv) return;

	glBegin( GL_TRIANGLES);
	int act_tex = -1;
	const int istop = int( f.size());
	if( !use_normals && !use_texture) {
	// Render faces only (no normals or textures)
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			for( int i = 0; i < istop; i++) {
				glVertex3fv( tv + (int( f[i].vid[0]) << 4));
				glVertex3fv( tv + (int( f[i].vid[1]) << 4));
				glVertex3fv( tv + (int( f[i].vid[2]) << 4));
			}
			break;
		}

	} else if( use_normals && !use_texture) {
	// Render faces and normals (no textures)
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			for( int i = 0; i < istop; i++) {
				glNormal3fv( (float *)&(f[i].n[0]));
				glVertex3fv( tv + (int( f[i].vid[0]) << 4));

				glNormal3fv( (float *)&(f[i].n[0]));
				glVertex3fv( tv + (int( f[i].vid[1]) << 4));

				glNormal3fv( (float *)&(f[i].n[0]));
				glVertex3fv( tv + (int( f[i].vid[2]) << 4));
			}
			break;
		}

	} else if( !use_normals && use_texture) {
		rgFloat *tv_ptr;
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			for( int i = 0; i < istop; i++) {
				HLP_BIND_TEXTURE_IF_NEEDED( i);
				tv_ptr = tv + (int( f[i].vid[0]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[i].vid[1]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[i].vid[2]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glVertex3fv( tv_ptr);
			}
			break;
		}

	} else {
		rgFloat *tv_ptr;
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			for( int i = 0; i < istop; i++) {
				HLP_BIND_TEXTURE_IF_NEEDED( i);
				tv_ptr = tv + (int( f[i].vid[0]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glNormal3fv( tv_ptr + 4);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[i].vid[1]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glNormal3fv( tv_ptr + 4);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[i].vid[2]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glNormal3fv( tv_ptr + 4);
				glVertex3fv( tv_ptr);
			}
			break;
		}

	}
	glEnd();
}

void sk3DRomModel::render_faces( bool use_normals, bool use_texture,
	int *flist, const int count)
{
	if( !tv) return;

	glBegin( GL_TRIANGLES);
	int act_tex = -1;
	if( !use_normals && !use_texture) {
	// Render faces only (no normals or textures)
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			for( int i = 0; i < count; i++) {
				glVertex3fv(
					tv + (int( f[*flist].vid[0]) << 4));
				glVertex3fv(
					tv + (int( f[*flist].vid[1]) << 4));
				glVertex3fv(
					tv + (int( f[*flist].vid[2]) << 4));

				flist++;
			}
			break;
		}

	} else if( use_normals && !use_texture) {
	// Render faces and normals (no textures)
		rgFloat *tv_ptr;
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			for( int i = 0; i < count; i++) {
				tv_ptr = tv + (int( f[*flist].vid[0]) << 4);
				glNormal3fv( tv_ptr + 4);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[*flist].vid[1]) << 4);
				glNormal3fv( tv_ptr + 4);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[*flist].vid[2]) << 4);
				glNormal3fv( tv_ptr + 4);
				glVertex3fv( tv_ptr);

				flist++;
			}
			break;
		}

	} else if( !use_normals && use_texture) {
		rgFloat *tv_ptr;
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			for( int i = 0; i < count; i++) {
				HLP_BIND_TEXTURE_IF_NEEDED( *flist);
				tv_ptr = tv + (int( f[*flist].vid[0]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[*flist].vid[1]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[*flist].vid[2]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glVertex3fv( tv_ptr);

				flist++;
			}
			break;
		}

	} else {
		rgFloat *tv_ptr;
		switch( render_format) {
		case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
			for( int i = 0; i < count; i++) {
				HLP_BIND_TEXTURE_IF_NEEDED( *flist);
				tv_ptr = tv + (int( f[*flist].vid[0]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glNormal3fv( tv_ptr + 4);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[*flist].vid[1]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glNormal3fv( tv_ptr + 4);
				glVertex3fv( tv_ptr);

				tv_ptr = tv + (int( f[*flist].vid[2]) << 4);
				glTexCoord2fv( tv_ptr + 8);
				glNormal3fv( tv_ptr + 4);
				glVertex3fv( tv_ptr);

				flist++;
			}
			break;
		}

	}
	glEnd();
}
*/
#undef HLP_BIND_TEXTURE_IF_NEEDED

void sk3DRomModel::prepare_texture_coords_for_rendering()
{
/*
	if( !tv) return;
	const int istop = int( f.size());
	int i;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		
		break;
	}
*/
}

void sk3DRomModel::prepare_normals_for_rendering()
{
/*
	if( !tv) return;
	const int istop = int( f.size());
	int i;
	switch( render_format) {
	case SK3DROM_RENDER_FORMAT_V4N3E1T2E6:
		break;
	}
*/
}
/*
void sk3DRomModel::gen_optimal_face_render_order()
{
	if( ofro) {
		delete [] ofro;
		ofro = 0;
	}
	if( f.empty()) return;

	ofro = new int[ f.size()];

	// Render faces in texture groups, starting from the first texture
	// all the way up to the last texture.
	const int istop = int( f.size());
	const int jstop = int( );
	for( int i = 0; i < istop; i++) {
		
	}
}

*/


void sk3DRomModel::clear_bsp_data()
{
	bsp_bone_id.clear();
	while( !(bsp_tree.empty())) {
		if( bsp_tree[0]) delete (bsp_tree[0]);
		bsp_tree.erase( bsp_tree.begin());
	}
}

// Description: Builds a BSP tree containing "alpha" and "bsp" faces. Faces
//	making use of any alpha component are "alpha" faces and any faces
//	which have their "bsp" bit set are bsp faces.
void sk3DRomModel::compile_alpha_faces()
{
	// Delete the current BSP information.
	clear_bsp_data();

	// If there are no faces then there is nothing to do.
	if( f.empty()) return;

	// Alpha faces are generated from the "skout" skeleton.
	// FIXME: This will destroy the the current state of the "tv" array.
	//	perhaps it could be backed up or another method could be used
	//	to get the vertices in the BSP's desired transformation.
	transform_skel_frame();

	// Scan all faces and mark them as either alpha or non-alpha faces.
	// Also, build a list of all alpha faces that will be part of a BSP
	// tree.
	std::vector <rgBSPFace <rgBSP_NTTC_ExtData> > bspFace;
	std::vector <int> bone_id;
	for( int i = 0; i < int( f.size()); i++) {
		
		bool is_alpha = false;

		// If this face makes use of the alpha colour component, then
		// it is an alpha face (only check alpha color component if
		// the face makes use of per vertex color).
		if( f[i].use_color()) {
			if(	((f[i].li[3] != 0xFF) || (f[i].li[7] != 0xFF)
				|| (f[i].li[11] != 0xFF))) is_alpha = true;
		}

		// If the face does not make use of the alpha colour component
		// but does make use of an alpha enabled texture, then it is an
		// alpha face.
		sk3DRomTexture *tex = 0;
		if( f[i].tid >= 0) tex = &(texture[ f[i].tid]);
		if( (!is_alpha) && tex && (tex->is_alpha())) is_alpha = true;

		// If the face is an alpha face, then mark it as such and add
		// it to the list of BSP faces.
		if( is_alpha || f[i].is_bsp()) {
			f[i].flags |= SK3DROM_FACE_ALPHA_BIT;
			f[i].set_bsp( true);

			// Add the face to the BSP list.
			bspFace.push_back(
				rgBSPFace <rgBSP_NTTC_ExtData>());
			rgBSPFace <rgBSP_NTTC_ExtData> &tmp_face =
				bspFace[ bspFace.size() - 1];
			#define TMP_ADD_VERTEX( mv_v) \
				tmp_face.v.push_back( \
					rgBSPVertex <rgBSP_NTTC_ExtData> ( \
					(mv_v)))

			TMP_ADD_VERTEX( get_tv_vertex( f[i].vid[0]));
			TMP_ADD_VERTEX( get_tv_vertex( f[i].vid[1]));
			TMP_ADD_VERTEX( get_tv_vertex( f[i].vid[2]));
			tmp_face.gen_normal();

			if( tex) tmp_face.m_tid = tex->texture_id();

			// Set the per-vertex normals.
			tmp_face.v[0].data.n = f[i].n[0];
			tmp_face.v[1].data.n = f[i].n[1];
			tmp_face.v[2].data.n = f[i].n[2];

			// Set the per-vertex texture coordinates (unit 1).
			tmp_face.v[0].data.u1 = f[i].map[0];
			tmp_face.v[0].data.v1 = f[i].map[1];
			tmp_face.v[1].data.u1 = f[i].map[2];
			tmp_face.v[1].data.v1 = f[i].map[3];
			tmp_face.v[2].data.u1 = f[i].map[4];
			tmp_face.v[2].data.v1 = f[i].map[5];

			// Set the per-vertex colour.
			// Use a temporary color converter.
			#define TMPCC(n) (float( n) / 255.0f)
			tmp_face.v[0].data.r = TMPCC( f[i].li[0]);	// R
			tmp_face.v[0].data.g = TMPCC( f[i].li[1]);	// G
			tmp_face.v[0].data.b = TMPCC( f[i].li[2]);	// B
			tmp_face.v[0].data.a = TMPCC( f[i].li[3]);	// A

			tmp_face.v[1].data.r = TMPCC( f[i].li[4]);	// R
			tmp_face.v[1].data.g = TMPCC( f[i].li[5]);	// G
			tmp_face.v[1].data.b = TMPCC( f[i].li[6]);	// B
			tmp_face.v[1].data.a = TMPCC( f[i].li[7]);	// A

			tmp_face.v[2].data.r = TMPCC( f[i].li[8]);	// R 
			tmp_face.v[2].data.g = TMPCC( f[i].li[9]);	// G
			tmp_face.v[2].data.b = TMPCC( f[i].li[10]);	// B
			tmp_face.v[2].data.a = TMPCC( f[i].li[11]);	// A
			#undef TMPCC

			bone_id.push_back( hlp_get_face_bone( f[i], skel));
	
			#undef TMP_ADD_VERTEX
			
		} else {
			f[i].flags &= ~SK3DROM_FACE_ALPHA_BIT;
			f[i].flags &= ~SK3DROM_FACE_BSP_BIT;
		}
	}

	// If there are no alpha faces, then no BSP list needs to be built.
	if( bspFace.empty()) return;


	// FIXME: The below algorthim is not efficient, but it works.
	
	// All BSP faces must now be seperated into seperate BSP lists based
	// on bones. That is, all faces with two or more of their verticies
	// controlled by a bone will be added together in their own BSP list.
	// Faces which do not have at least two controlled verticies will be
	// part of a final "non-bone" BSP list.
	hlp_seperate_BSP_into_BSP_bone_lists( *this, bspFace, bone_id);
}

int hlp_get_face_bone( sk3DRomFace &f, sk3DRomSkel &s)
{
	// At least two vertices from the face "f" must be part of a
	// bone's vertex list "vid" to be a part of that bone. First bone
	// found matching this criteria is the controlling bone (the bone
	// that control's the face.
	for( int i = 0; i < int( s.b.size()); i++) {

		int count = 0;
		for( int j = 0; j < int( s.b[i].vid.size()); j++) {
			if(	(f.vid[0] == s.b[i].vid[j]) ||
				(f.vid[1] == s.b[i].vid[j]) ||
				(f.vid[2] == s.b[i].vid[j])) {

				count++;
				if( 2 == count) return i;
			}
		}
	}

	// The face is not controlled by any bone (it is boneless).
	return -1;
}

void hlp_bind_bsp_vertices(
	std::vector <rgBSPVertex <rgBSP_NTTC_ExtData> *> &v, void *p_data)
{
	if( !p_data) return;
	rgCoordSystem *cs = (rgCoordSystem *)p_data;
	for( int i = 0; i < int( v.size()); i++)
		(v[i])->v = cs->change_point_csys( (v[i])->v);
}

void hlp_bind_bsp_faces_to_bones( const sk3DRomSkel &s,
	std::vector < rgBSPTree <rgBSP_NTTC_ExtData> *> &bsp_tree,
	std::vector <int> &bsp_bone_id)
{
	for( int i = 0; i < int( bsp_bone_id.size()); i++) {
		if( bsp_bone_id[i] >= 0) {
			rgCoordSystem cs( s.get_full_bone_jcs(
				bsp_bone_id[i]));
			(bsp_tree[i])->transform_node_verts(
				hlp_bind_bsp_vertices, &cs);
		}
	}
}

void hlp_seperate_BSP_into_BSP_bone_lists( sk3DRomModel &mdl,
	std::vector <rgBSPFace <rgBSP_NTTC_ExtData> > &bspFace,
	std::vector <int> &bone_id)
{
	// NOTE: "mdl.bsp_tree" would be cleared of all BSP lists already.

	std::vector < std::vector <rgBSPFace <rgBSP_NTTC_ExtData> > > blist;
	std::vector <int> target_face;

	// NOTE: "i" starts at -1. This is because "i" is never used to
	//	directly index bones. It is simply used to identify them. -1
	//	means "bone-less" or not controlled by a bone.
	int j;
	for( int i = -1; i < int( mdl.skel.b.size()); i++) {

		// Find all faces that are controlled by this bone.
		for( j = 0; j < int( bone_id.size()); j++)
			if( bone_id[j] == i) target_face.push_back( j);

		// If there is at least one bone in the list, then create a
		// BSP face list for this bone.
		if( !target_face.empty()) {
			// Identify this BSP bone list as being for bone "i".
			mdl.bsp_bone_id.push_back( i);

			// Create the BSP bone list.
			blist.push_back( std::vector
				<rgBSPFace <rgBSP_NTTC_ExtData> >());
			std::vector <rgBSPFace <rgBSP_NTTC_ExtData> > &tmp_face
				= blist[ blist.size() - 1];
			
			// Add all of the target faces to the BSP bone list.
			while( !(target_face.empty())) {
				tmp_face.push_back( bspFace[ target_face[0]]);
				target_face.erase( target_face.begin());
			}
		}
	}

	bone_id.clear();
	bspFace.clear();


	// Now build BSP tree's for each BSP bone list.
	while( !(blist.empty())) {
		rgBSPTree <rgBSP_NTTC_ExtData> *p =
			new rgBSPTree <rgBSP_NTTC_ExtData> ();
		mdl.bsp_tree.push_back( p);
		std::cerr << "splits = "
			<< p->gen_tree( blist[0], rgBSP_NTTC_interp_fn)
			<< "\n";
		p->convex_hull_renderer = rgBSP_NTTC_convex_hull_renderer;

		blist.erase( blist.begin());
	}

	// Rebind all resulting faces to their respective bones.
	hlp_bind_bsp_faces_to_bones( mdl.skel, mdl.bsp_tree, mdl.bsp_bone_id);
}

void sk3DRomModel::mirror_tv_vertices( const int *vlist, int count,
	const rgVector origin, const rgVector normal,
	const rgVector mirror_vec)
{
	if( !(vlist) || (count <= 0)) return;

	rgVector v;
	rgVector u;
	while( count--) {
		v = get_tv_vertex( *vlist);
		u = rgMakePlanar( v - origin, normal);
		u = (u.dot( mirror_vec) * mirror_vec);
		set_tv_vertex( *vlist, v - (2.0f * u));
		vlist++;
	}
}

void sk3DRomModel::invert_mesh( const int *vlist, const int count)
{
	if( !(vlist) || (count < 3)) return;

	int vid[3];
	for( int i = 0; i < int( f.size()); i++) {
		if( triangle_in_mesh( i, vlist, count, vid)) {
			int tmp = f[i].vid[0];
			f[i].vid[0] = f[i].vid[2];
			f[i].vid[2] = tmp;
		}
	}
}

void sk3DRomModel::use_shared_vertex_array( const bool b)
{
	if( !m_gl_vertex_array) {
		if( !(m_gl_vertex_array = new sk3DRomVNTArray())) return;
	}
	m_gl_vertex_array->set_share_vertex( b);
	m_gl_vertex_array->create_arrays( *this);
	m_gl_vertex_array->update_arrays( *this);
}

int sk3DRomModel::create_vertex_array()
{
	delete_vertex_array();
	if( !(m_gl_vertex_array = new sk3DRomVNTArray())) return -1;
	m_gl_vertex_array->create_arrays( *this);
	m_gl_vertex_array->update_arrays( *this);
	return 0;
}

void sk3DRomModel::update_vertex_array()
{
	if( !tv) return;
	if( !m_gl_vertex_array) return;
	m_gl_vertex_array->update_arrays( *this);
}

void sk3DRomModel::render_vertex_array( sk3DRomDrawContext *dc)
{
	if( !m_gl_vertex_array) return;
	if( !dc) dc = &(skel.m_draw_context);
	m_gl_vertex_array->render_arrays();
}

void sk3DRomModel::delete_vertex_array()
{
	if( m_gl_vertex_array) {
		delete m_gl_vertex_array;
		m_gl_vertex_array = 0;
	}
}
