/*
Copyright (C) 2000-2005 Amour Hassan
                                                                                
This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.
                                                                                
This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
                                                                                
You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
                                                                                
*/

#ifndef RGOPENGL_H
#define RGOPENGL_H

#ifdef WINDOWS_OS
#include <windows.h>
#endif
#include <GL/gl.h>
#include <rgStandard.h>
#include <rgBMP.h>
#include <vector>

#define MAX_RGGLTEXTURES_WITH_REFERENCES	2048

#define RGGL_MAX_TEX_WIDTH 256
#define RGGL_MAX_TEX_HEIGHT 256

/***************************************************************************
 * Flags for rgGLTexture::load().
 ***************************************************************************/
#define RGGL_SMOOTH_TEXTURE	0x00000001
#define RGGL_RGB_ALPHA_EXPAND	0x00000002	/* Create an alpha channel
						 * from the image file's RGB
						 * components (per pixel alpha
						 * conversion).
						 */
#define RGGL_MIPMAP		0x00000004
#define RGGL_COLOR_KEY		0x00000008
#define RGGL_LARGE_TEXTURE	0x00000010	/* Allows textures greater
						 * than 256 x 256 to be loaded.
						 */

#define RG_RGB( r, g, b) ((GLuint)(((r) << 24) | ((g) << 16) | ((b) << 8)))
#define RG_RGBA( r, g, b, a) \
	((GLuint)(((r) << 24) | ((g) << 16) | ((b) << 8) | (a)))

class rgGLTexture_ref_tag
{
public:
	GLuint id;
	int count;

// FIXME: The Visual C++ compiler wants a default constructor. Why!!? A
//	default constructor should never be used for this class. This
//	default constructor is untested and Unix-like systems do not use it.
//	This may cause programs to crash, but fortunately it will be restricted
//	to just windows systems.
//#ifdef WINDOWS_OS
//	rgGLTexture_ref_tag(){ id = 0; count = -1;}
//#endif

	rgGLTexture_ref_tag( const int p_id, const int p_count) {
		id = p_id;
		count = p_count;
	}
};

class rgGLTexture
{
public:
	
	rgGLTexture();

	rgGLTexture( const rgGLTexture &t);

	/*
	 * Description: Destroy's the loaded texture (if any). Any texture
	 *	"name" (ID number) used will be given back to OpenGL so that
	 *	it can be used again for later textures.
	 */
	~rgGLTexture();

	rgGLTexture &operator =( const rgGLTexture &t);

	/* Makes this texture the active texture in OpenGL (that is, this
	 * function calls glBindTexture( textureID)). If no valid texture
	 * exists when bind is called, then texture ID 0 is bound to OpenGL
	 * (this makes it so that no texture is bound).
	 */
	void bind() const { glBindTexture( GL_TEXTURE_2D, textureID);}

	/*
	 * Description: Deletes the currently loaded texture and gives the
	 *	texture "name" (ID number) back to OpenGL.
	 */
	void clear();

	int getWidth() const { return width;}
	int getHeight() const { return height;}
	
	/*
	 * Returns: RGERR_SUCCESS is returned if the texture was loaded. An
	 *	error value is returned otherwise.
	 */
	int load( const char *fname);
	int load( const char *fname, int flags);

	int load_sub_image( rgBMP &bmp, int x, int y, int w, int h,
		int flags);

	/*
	 * Returns: The OpenGL texture "name" for the currently loaded
	 *	texture is returned. If no texture was loaded, then 0 is
	 *	returned.
	 */
	GLuint getTextureID() const { return textureID;}

	void setColorKey( int r, int g, int b);
private:
	int width;
	int height;

	GLuint colorKey;
	GLuint textureID;

	void genAlphaFromRGB( rgBMP &bmp);
	void genAlphaFromColorKey( rgBMP &bmp);

	bool find_texture_ref( const GLuint id, int *i);
	void add_texture_reference( const GLuint id);
	void inc_texture_reference( const GLuint id);
	void dec_texture_reference( const GLuint id);
	static std::vector <rgGLTexture_ref_tag> generated_textures;

	void copy_data( const rgGLTexture &t);


	//void reference_rgGLTexture( rgGLTexture *t) const;
};

class rgGLLight
{
public:
	int id;

	rgFloat pos[4];
	rgFloat amb[4];
	rgFloat dif[4];
	rgFloat spec[4];

	rgGLLight();

	void set_pos() const { glLightfv( id, GL_POSITION, pos);}
	void set_amb() const { glLightfv( id, GL_AMBIENT, amb);}
	void set_dif() const { glLightfv( id, GL_DIFFUSE, dif);}
	void set_spec() const { glLightfv( id, GL_SPECULAR, spec);}

	void set_pos( rgFloat x, rgFloat y, rgFloat z, rgFloat w);
	void set_amb( rgFloat r, rgFloat b, rgFloat g, rgFloat a);
	void set_dif( rgFloat r, rgFloat b, rgFloat g, rgFloat a);
	void set_spec( rgFloat r, rgFloat b, rgFloat g, rgFloat a);

	void enable() const { glEnable( id);}
	void disable() const { glDisable( id);}
};


class rgGLCamera
{
public:
	rgCoordSystem cs;

	rgGLCamera();
	rgGLCamera( rgVector p, rgVector p_look_at, rgVector up);

	void apply() const;
	void look( rgVector eye, rgVector up);
	void look_at( rgVector p);
	void look_at( rgVector p, rgVector up);
	void move_to( rgVector p, const float w) {
		cs.p += (p - cs.p) * w;
	}

	rgVector calc_side_vec() const { return cs.up.cross( cs.eye);}
};

#define RGGL_VIEW_MATRIX_SRT	0
#define RGGL_VIEW_MATRIX_STR	1
#define RGGL_VIEW_MATRIX_RTS	2
#define RGGL_VIEW_MATRIX_TRS	3
class rgGLViewMatrix
{
public:
	rgMatrix camM;
	rgMatrix camT;
	rgMatrix camR;
	rgMatrix camS;

	// Inverse matricies
	rgMatrix icamM;
	rgMatrix icamT;
	rgMatrix icamR;
	rgMatrix icamS;

	rgGLViewMatrix();

	void reset();
	void rotate( const rgFloat ang, rgVector axis);
	void set_transformation_order( int order);
	void scale( rgFloat x, rgFloat y, rgFloat z);
	void translate( rgVector v);
	void translate( rgFloat x, rgFloat y, rgFloat z);
	void update();
	void update( bool tm, bool rm, bool sm);

	// Description: Performs a 3-way (pitch, roll, yaw) rotation. The
	//	location of the point "src" determines which type of rotation
	//	will be done. The distance of "des" from "src" determines the
	//	magnitude of the rotation.
	//
	// Parameters:
	//	width - Width of the window.
	//	height - Height of the window.
	//	src - 2D point in window coordinates (initial point).
	//	des - 2D point in window coordinates (final point).
	//	m - Rotational matrix is saved here.
	//	im - Inverse of rotational matrix is saved here.
	//	set - If "true" rotation is applied to "this" view matrix,
	//		otherwise the rotation results are not applied.
	void do_view_3way_rotate( const float width, const float height,
		const rgVector src, const rgVector des, rgMatrix *m,
		rgMatrix *im, const bool set);

	void do_view_4way( const float width, const float height,
		const rgVector src, const rgVector des, rgMatrix *m,
		rgMatrix *im, const bool set);

	void do_view_rotate( const rgVector forward, const rgVector src,
		const rgVector des, const rgVector axis, rgMatrix *m,
		rgMatrix *im, const bool set);

	void do_view_circular_rotate( const float width, const float height,
		const rgVector src, const rgVector des, const rgVector axis,
		rgMatrix *m, rgMatrix *im, const bool set);

	void do_view_translate( const float width, const float height,
		const rgVector src, const rgVector des, rgMatrix *m,
		rgMatrix *im, const bool set);

	void do_view_scale( const float height, const rgVector src,
		rgVector des, const float k, rgMatrix *m, rgMatrix *im,
		const bool set);
private:
	int trans_order;
};

void loadRGMatrix( const rgMatrix &m);
void getRGMatrix( rgMatrix &m);

int rggl_read_pixels( rgBMP &bmp, const GLint buf, const int x, const int y,
        const int width, const int height);

void rggl_get_ortho_dimensions( float &w, float &h, rgMatrix &m);
void rggl_ortho_window_to_world( const float w, const float h, rgVector *v,
	const int count, const rgMatrix *inv);
void rggl_ortho_world_to_window( const float w, const float h, rgVector *v,
	int count);

void rggl_print_error();
void rggl_print_error( std::ostream &o);

inline void glVertex( const rgVector &v) { glVertex3f( v.x, v.y, v.z);}
inline void glNormal( const rgVector &n) { glNormal3f( n.x, n.y, n.z);}
inline void glColor3( const rgVector &c) { glColor3f( c.x, c.y, c.z);}

rgVector rggl_unproject( float x, float y, float z);
//void rggl_pick_ray( int x, int y, float fov, rgVector &rp, rgVector &rd);

#endif
